# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Thing',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('big', models.BooleanField(default=False)),
                ('clever', models.BooleanField(default=False)),
                ('element_type', models.CharField(max_length=1, choices=[(b'e', b'Earth'), (b'w', b'Water'), (b'a', b'Air'), (b'f', b'Fire')])),
                ('count', models.IntegerField(default=0)),
                ('description', models.TextField(blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



from django.core.urlresolvers import reverse
from selenium.common.exceptions import NoSuchElementException, TimeoutException

from django_functest.exceptions import (
    SeleniumCantUseElement, WebTestCantUseElement, WebTestMultipleElementsException, WebTestNoSuchElementException
)
from django_functest.tests.models import Thing

from .base import ChromeBase, FirefoxBase, PhantomJSBase, WebTestBase


class TestCommonBase(object):
    def setUp(self):
        super(TestCommonBase, self).setUp()
        self.thing = Thing.objects.create(name="Rock",
                                          big=True,
                                          clever=False,
                                          element_type=Thing.ELEMENT_EARTH,
                                          count=1,
                                          description="Hard thing")

    def test_get_url(self):
        self.get_url('admin:login')
        url = self.current_url
        self.assertTrue(url.endswith("/admin/login/"))
        self.assertTrue(url.startswith("http://"))

    def test_get_literal_url(self):
        url = reverse('admin:login')
        self.get_literal_url(url)
        self.assertUrlsEqual(url)

    def test_assertUrlsEqual_default(self):
        self.get_url('admin:login')
        self.assertRaises(AssertionError, lambda: self.assertUrlsEqual("foo"))
        self.assertUrlsEqual("/admin/login/")

    def test_assertUrlsEqual_path(self):
        self.assertRaises(AssertionError, lambda: self.assertUrlsEqual("/login/", "/admin/login/"))
        self.assertUrlsEqual("/login/", "/login/")

    def test_assertUrlsEqual_query(self):
        self.assertRaises(AssertionError, lambda: self.assertUrlsEqual("/foo/?q=1", "/foo/"))
        self.assertRaises(AssertionError, lambda: self.assertUrlsEqual("/foo/?q=1", "/foo/?q=2"))
        self.assertUrlsEqual("/foo/?q=1", "/foo/?q=1")

    def test_assertUrlsEqual_host(self):
        self.assertUrlsEqual("/foo/", "//example.com/foo/")
        self.assertUrlsEqual("//example.com/foo/", "//example.com/foo/")
        self.assertRaises(AssertionError, lambda: self.assertUrlsEqual("//example.com/foo/",
                                                                       "//other.com/foo/"))

    def test_assertUrlsEqual_protocol(self):
        self.assertUrlsEqual("http://example.com/foo/", "//example.com/foo/")
        self.assertUrlsEqual("http://example.com/foo/", "http://example.com/foo/")
        self.assertRaises(AssertionError, lambda: self.assertUrlsEqual("http://example.com/foo/",
                                                                       "https://example.com/foo/"))

    def test_assertTextPresent(self):
        self.get_url('django_functest.test_misc')
        self.assertTextPresent("Hello world")
        # Check escaping
        self.assertTextPresent("from 'me' & \"friends\"")
        self.assertRaises(AssertionError, lambda: self.assertTextPresent("Something definitely not there"))

    def test_assertTextAbsent(self):
        self.get_url('django_functest.test_misc')
        self.assertTextAbsent("Something definitely not there")
        self.assertRaises(AssertionError, lambda: self.assertTextAbsent("Hello world"))
        self.assertRaises(AssertionError, lambda: self.assertTextAbsent("from 'me' & \"friends\""))

    def test_current_url(self):
        self.get_url('admin:login')
        # Check it really is a full URL
        self.assertTrue(self.current_url.startswith('http'))

    def test_is_element_present(self):
        self.get_url('admin:login')
        self.assertTrue(self.is_element_present('#id_username'))
        self.assertFalse(self.is_element_present('#id_something_not_there'))

    def refresh_thing(self):
        self.thing = Thing.objects.get(id=self.thing.id)
        return self.thing

    def test_fill(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.fill({'#id_name': "New name",
                   '#id_big': False,
                   '#id_clever': True,
                   '#id_element_type': Thing.ELEMENT_AIR,
                   '#id_count': 5,
                   '#id_description': "Soft thing\r\nwith line breaks",
                   })
        self.submit('input[name=change]')
        self._assertThingChanged()

    def test_fill_by_id(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.fill_by_id({'id_name': "New name",
                         'id_big': False,
                         'id_clever': True,
                         'id_element_type': Thing.ELEMENT_AIR,
                         'id_count': 5,
                         'id_description': "Soft thing\r\nwith line breaks",
                         })
        self.submit('input[name=change]')
        self._assertThingChanged()

    def test_fill_by_name(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.fill_by_name({'name': "New name",
                           'big': False,
                           'clever': True,
                           'element_type': Thing.ELEMENT_AIR,
                           'count': 5,
                           'description': "Soft thing\r\nwith line breaks",
                           })
        self.submit('input[name=change]')
        self._assertThingChanged()

    def test_fill_by_text(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.fill_by_text({'#id_element_type': 'Water'})
        self.submit('input[name=change]')
        self.refresh_thing()
        self.assertEqual(self.thing.element_type,
                         Thing.ELEMENT_WATER)

    def test_fill_by_text_missing(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.assertRaises(self.TextNotFoundException, lambda: self.fill_by_text({'#id_element_type': 'Plasma'}))

    def test_fill_by_text_for_unsupported(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.assertRaises(self.ElementUnusableException, lambda: self.fill_by_text({'#id_count': 'Water'}))

    def _assertThingChanged(self):
        thing = self.refresh_thing()
        self.assertEqual(thing.name, "New name")
        self.assertEqual(thing.big, False)
        self.assertEqual(thing.clever, True)
        self.assertEqual(thing.element_type, Thing.ELEMENT_AIR)
        self.assertEqual(thing.count, 5)
        self.assertEqual(thing.description, "Soft thing\r\nwith line breaks")

    def test_fill_no_element_error(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.assertRaises(self.ElementNotFoundException, lambda: self.fill({'#id_blahblah': "New name"}))

    def test_submit(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.submit('button[name=clear]')
        thing = self.refresh_thing()
        self.assertEqual(thing.name, "")

    def test_follow_link(self):
        self.get_url('list_things')
        self.follow_link('a.edit')
        self.assertUrlsEqual(reverse('edit_thing', kwargs={'thing_id': self.thing.id}))

    def test_follow_link_not_found(self):
        self.get_url('list_things')
        self.assertRaises(self.ElementNotFoundException, lambda: self.follow_link('a.foobar'))

    def test_back(self):
        self.get_url('list_things')
        self.follow_link('a.edit')
        self.back()
        self.assertUrlsEqual(reverse('list_things'))

    def test_multiple_back(self):
        # We could test the behaviour regarding forms, especially those that
        # submit to the same URL and then redirect to the same URL. However,
        # Firefox and Chrome behave differently here - Firefox produces
        # fewer history entries.
        self.get_url('list_things')
        self.follow_link('a.edit')
        edit_url = reverse('edit_thing', kwargs={'thing_id': self.thing.id})
        self.assertUrlsEqual(edit_url)
        self.submit('button[name=clear]')
        self.assertUrlsEqual(reverse('thing_cleared', kwargs={'thing_id': self.thing.id}))
        self.assertTextPresent("was cleared")
        self.back()
        self.assertUrlsEqual(edit_url)
        self.back()
        self.assertUrlsEqual(reverse('list_things'))

    def test_set_session_data(self):
        self.set_session_data({'name': 'The Jabberwocky'})
        self.get_url('django_functest.test_misc')
        self.assertTextPresent("Hello to The Jabberwocky")

    def test_value(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.assertEqual(self.value('#id_name'),
                         "Rock")
        self.assertEqual(self.value('#id_big'),
                         True)
        self.assertEqual(self.value('#id_clever'),
                         False)
        self.assertEqual(self.value('#id_element_type'),
                         'e')
        self.assertEqual(self.value('#id_description'),
                         'Hard thing')


class TestFuncWebTestCommon(TestCommonBase, WebTestBase):

    ElementNotFoundException = WebTestNoSuchElementException
    TextNotFoundException = ValueError
    ElementUnusableException = WebTestCantUseElement

    def test_is_full_browser_attribute(self):
        self.assertEqual(self.is_full_browser_test, False)

    def test_fill_multiple_matches(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.assertRaises(WebTestMultipleElementsException, lambda: self.fill({'input[type=checkbox]': True}))

    def test_fill_element_without_name(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.assertRaises(WebTestCantUseElement, lambda: self.fill({'#id_badinput1': "Hello"}))

    def test_fill_element_outside_form(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.assertRaises(WebTestCantUseElement, lambda: self.fill({'#id_badinput2': "Hello"}))

    def test_submit_no_auto_follow(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.submit('input[name=change]', auto_follow=False)
        self.assertEqual(self.last_response.status_int, 302)

    def test_follow_link_multiple_matches(self):
        Thing.objects.create(name="Another")
        self.get_url('list_things')
        self.assertRaises(WebTestMultipleElementsException, lambda: self.follow_link('a.edit'))

    def test_follow_link_no_href(self):
        self.get_url('list_things')
        self.assertRaises(WebTestCantUseElement, lambda: self.follow_link('a.javascriptonly'))

    def test_get_literal_url_auto_follow(self):
        url = '/admin/login'  # No trailing '/'
        self.get_literal_url(url, auto_follow=True)
        self.assertUrlsEqual(url + '/')

        self.get_literal_url(url, auto_follow=False)
        self.assertUrlsEqual(url)
        self.assertEqual(self.last_response.status_int, 301)

    def test_get_literal_url_expect_errors(self):
        url = '/a_404_url/'
        self.get_literal_url(url, expect_errors=True)
        self.assertEqual(self.last_response.status_int, 404)
        self.assertRaises(Exception, lambda: self.get_literal_url(url, expect_errors=False))


class TestFuncSeleniumCommonBase(TestCommonBase):

    ElementNotFoundException = TimeoutException
    TextNotFoundException = NoSuchElementException
    ElementUnusableException = SeleniumCantUseElement

    def test_is_full_browser_attribute(self):
        self.assertEqual(self.is_full_browser_test, True)

    def test_fill_with_scrolling(self):
        url = reverse('edit_thing', kwargs=dict(thing_id=self.thing.id)) + "?add_spacers=1"
        self.get_literal_url(url)
        self.fill({'#id_name': "New name",
                   '#id_big': False,
                   '#id_clever': True,
                   '#id_element_type': Thing.ELEMENT_AIR,
                   '#id_count': 5,
                   '#id_description': "Soft thing\r\nwith line breaks",
                   })
        self.submit('input[name=change]')
        self._assertThingChanged()

    def test_submit_no_wait_for_reload(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.submit('button[name=check]', wait_for_reload=False)
        self.assertTextPresent("Everything is fine")

    def test_submit_slow_page(self):
        url = reverse('edit_thing', kwargs=dict(thing_id=self.thing.id)) + "?add_js_delay=5"
        self.get_literal_url(url)
        self.fill({'#id_name': "New name",
                   '#id_big': False,
                   '#id_clever': True,
                   '#id_element_type': Thing.ELEMENT_AIR,
                   '#id_count': 5,
                   '#id_description': "Soft thing\r\nwith line breaks",
                   })
        self.submit('input[name=change]')
        self._assertThingChanged()


class TestFuncSeleniumCommonFirefox(TestFuncSeleniumCommonBase, FirefoxBase):
    pass


class TestFuncSeleniumCommonChrome(TestFuncSeleniumCommonBase, ChromeBase):
    pass


class TestFuncSeleniumCommonPhantomJS(TestFuncSeleniumCommonBase, PhantomJSBase):
    pass



import os
import unittest

from django.contrib.auth import get_user_model
from django.core.urlresolvers import reverse
from selenium.common.exceptions import NoSuchElementException

from django_functest import AdminLoginMixin

from .base import ChromeBase, FirefoxBase, PhantomJSBase
from .models import Thing


# Tests for Selenium specific methods

class TestFuncSeleniumSpecificBase(AdminLoginMixin):

    def setUp(self):
        super(TestFuncSeleniumSpecificBase, self).setUp()
        self.thing = Thing.objects.create(name="Rock",
                                          big=True,
                                          clever=False,
                                          element_type=Thing.ELEMENT_EARTH)
        User = get_user_model()
        self.user = User.objects.create_superuser("admin", "admin@example.com", "password")

    def test_is_element_displayed(self):
        self.get_url('admin:login')
        self.assertTrue(self.is_element_displayed('#id_username'))
        self.assertFalse(self.is_element_displayed('#id_something_else'))
        self.execute_script("document.querySelector('#id_username').style.display = 'none';")
        self.assertFalse(self.is_element_displayed('#id_username'))

    def test_click(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.click('button[name=check]')
        self.assertTextPresent("Everything is fine")

    def test_click_xpath(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.click(xpath='//button[@name="check"]')
        self.assertTextPresent("Everything is fine")

    def test_click_text(self):
        self.get_url('list_things')
        self.click(text='Edit Rock', text_parent_id="id_thinglist", wait_for_reload=True)
        self.assertUrlsEqual(reverse('edit_thing', kwargs={'thing_id': self.thing.id}))

    def test_click_text_missing(self):
        self.get_url('list_things')
        self.assertRaises(NoSuchElementException,
                          lambda: self.click(text='Edit Fridge', wait_timeout=0))
        self.assertRaises(NoSuchElementException,
                          lambda: self.click(text='Edit Rock',
                                             text_parent_id='fribble',
                                             wait_timeout=0))

    def test_double_click(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.click('button[name=check]', double=True)
        self.assertTextPresent("Everything is really fine")

    def test_double_click_element_that_changes(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.click('[name=debounced]', double=True)
        self.assertTextPresent("Pressed x 1")

    def test_double_click_element_that_disappears(self):
        self.get_url('edit_thing', thing_id=self.thing.id)
        self.click('[name=disappears]', double=True)
        self.assertTextPresent("Pressed x 1")

    def test_execute_script(self):
        self.get_url('django_functest.test_misc')
        self.assertEqual(self.execute_script("return 1 + 1;"), 2)

    def test_execute_script_with_args(self):
        self.get_url('django_functest.test_misc')
        retval = self.execute_script("return arguments[0] + arguments[1];", 1, 2)
        self.assertEqual(retval, 3)

    def test_hover(self):
        self.get_url('django_functest.test_misc')
        get_style = "return document.defaultView.getComputedStyle(document.querySelector('#hoverable'))['font-style']"
        self.assertEqual(self.execute_script(get_style),
                         "normal")
        self.hover('#hoverable')
        self.assertEqual(self.execute_script(get_style),
                         "italic")

    def test_switch_window(self):
        self.do_login(username="admin", password="password")
        self.get_url("admin:auth_user_change", self.user.id)
        self.click('#add_id_groups')
        old_window, new_window = self.switch_window()
        self.fill({'#id_name': 'My new group'})
        self.switch_window(old_window)
        self.fill({'#id_first_name': 'My first name'})
        self.switch_window(new_window)
        self.submit('input[name=_save]', window_closes=True)
        self.switch_window(old_window)
        self.submit('input[name=_save]')

        User = get_user_model()
        user = User.objects.get(id=self.user.id)
        self.assertEqual([g.name for g in user.groups.all()],
                         ["My new group"])
        self.assertEqual(user.first_name, "My first name")

    def test_save_screenshot(self):
        testname = ("django_functest.tests.test_selenium.{0}.test_save_screenshot"
                    .format(self.__class__.__name__))

        fname = None
        try:
            fname = self.save_screenshot()
            self.assertIn(testname, fname)
            self.assertTrue(os.path.exists(fname))
        finally:
            if fname is not None:
                os.unlink(fname)


class TestFuncSeleniumSpecificFirefox(TestFuncSeleniumSpecificBase, FirefoxBase):

    # This fails on some Firefox versions, at least on Travis, but not locally,
    # so it is difficult to know how to write a better test.
    test_hover = unittest.expectedFailure(TestFuncSeleniumSpecificBase.test_hover)


class TestFuncSeleniumSpecificChrome(TestFuncSeleniumSpecificBase, ChromeBase):
    pass


class TestFuncSeleniumSpecificPhantomJS(TestFuncSeleniumSpecificBase, PhantomJSBase):

    # This fails for PhantomJS that is installed on Travis server, hard to debug
    # why.
    test_double_click_element_that_disappears = unittest.expectedFailure(TestFuncSeleniumSpecificBase.test_double_click_element_that_disappears)  # noqa


# Test class attribute `browser_window_size` works correctly:

class TestBrowserSizeBase(object):
    browser_window_size = (2800, 1400)

    def _get_window_size(self):
        if self._driver.name == "phantomjs":
            return self.execute_script("return [document.width, document.height]")
        else:
            return self.execute_script("return [window.outerWidth, window.outerHeight]")

    def test_size(self):
        width, height = self._get_window_size()
        self.assertTrue(2795 < width < 2805)
        self.assertTrue(1395 < height < 1405)

    def test_resize(self):
        self.set_window_size(400, 300)
        width, height = self._get_window_size()
        self.assertTrue(395 < width < 405)
        self.assertTrue(295 < height < 305)


class TestBrowserSizeFirefox(TestBrowserSizeBase, FirefoxBase):
    pass


class TestBrowserSizeChrome(TestBrowserSizeBase, ChromeBase):
    pass


class TestBrowserSizePhantomJS(TestBrowserSizeBase, PhantomJSBase):
    pass




class WebTestNoSuchElementException(Exception):
    pass


class WebTestMultipleElementsException(Exception):
    pass


class WebTestCantUseElement(Exception):
    pass


class SeleniumCantUseElement(Exception):
    pass



from .funcselenium import FuncSeleniumMixin
from .funcwebtest import FuncWebTestMixin
from .utils import AdminLoginMixin, ShortcutLoginMixin

__version__ = '0.1.1'

__all__ = ['FuncWebTestMixin', 'FuncSeleniumMixin', 'ShortcutLoginMixin', 'AdminLoginMixin']



import django
from django.conf.urls import patterns, url
from django.http import HttpResponse


def emptypage(request):
    # Minimal page needed for some tests
    return HttpResponse('<html><body></body></html>')


urlpatterns = [
    url(r'^__emptypage/$', emptypage, name='django_functest.emptypage'),
]

if django.VERSION < (1, 9):
    urlpatterns = patterns('', *urlpatterns)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

version = "0.1.1"

if sys.argv[-1] == 'publish':
    try:
        import wheel  # noqa
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()

if sys.argv[-1] == 'tag':
    print("Tagging the version on github:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

if sys.version_info < (3, 3):
    extra_requires = ['faulthandler>=2.4']
else:
    extra_requires = []


setup(
    name='django-functest',
    version=version,
    description="""Helpers for creating functional tests in Django, with a unified API for WebTest and Selenium tests.""",  # noqa
    long_description=readme + '\n\n' + history,
    author='Luke Plant',
    author_email='L.Plant.98@cantab.net',
    url='https://github.com/django-functest/django-functest',
    packages=[
        'django_functest',
    ],
    include_package_data=True,
    install_requires=[
        'django-webtest>=1.7.8',
        'selenium>=2.48.0',
        'PyVirtualDisplay>=0.1.2',
        'six>=1.10.0',
        'furl>=0.4.9',
        'pyquery>=1.2.10',
    ] + extra_requires,
    license="BSD",
    zip_safe=False,
    keywords='django-functest',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Framework :: Django :: 1.7',
        'Framework :: Django :: 1.8',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



from django import forms
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.utils.html import mark_safe

from .models import Thing


def test_misc(request):
    return render(request, "django_functest/tests/test_misc.html",
                  {'name': request.session.get('name', None)})


class ThingForm(forms.ModelForm):
    def __init__(self, add_spacers=False, **kwargs):
        super(ThingForm, self).__init__(**kwargs)
        self.add_spacers = add_spacers

    def as_p(self):
        retval = super(ThingForm, self).as_p()
        if self.add_spacers:
            # Hack to help test interacting with elements
            # that aren't in view.
            retval = mark_safe(retval.replace('</p>', '</p>' + ('<br>' * 100)))
        return retval

    class Meta:
        model = Thing
        fields = '__all__'


def edit_thing(request, thing_id):
    thing = Thing.objects.get(id=int(thing_id))
    add_spacers = 'add_spacers' in request.GET
    add_js_delay = int(request.GET.get('add_js_delay', '0'))

    if request.method == "POST":
        if 'clear' in request.POST:
            thing = Thing(id=thing.id)
            thing.save()
            return HttpResponseRedirect(reverse('thing_cleared', kwargs={'thing_id': thing_id}))
        else:
            thing_form = ThingForm(data=request.POST,
                                   instance=thing,
                                   add_spacers=add_spacers)
            if thing_form.is_valid():
                thing_form.save()
                return HttpResponseRedirect(reverse('edit_thing', kwargs={'thing_id': thing_id}))
    else:
        thing_form = ThingForm(instance=thing,
                               add_spacers=add_spacers)

    return render(request, "django_functest/tests/edit_thing.html",
                  {'thing_form': thing_form,
                   'thing': thing,
                   'add_js_delay': add_js_delay,
                   })


def list_things(request):
    return render(request, "django_functest/tests/list_things.html",
                  {'things': Thing.objects.all()})


def thing_cleared(request, thing_id):
    thing = Thing.objects.get(id=int(thing_id))
    return render(request, "django_functest/tests/thing_cleared.html",
                  {'thing': thing})



import logging
import os.path
import time
from datetime import datetime

from django.conf import settings
from django.core.urlresolvers import reverse
from pyvirtualdisplay import Display
from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException, NoSuchWindowException, StaleElementReferenceException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.support.ui import Select, WebDriverWait
from six import text_type

from .exceptions import SeleniumCantUseElement
from .utils import CommonMixin, get_session_store

logger = logging.getLogger(__name__)


def escape_selenium(text):
    # Selenium seems to do something strange with its page source function:
    #  &quot; gets converted back to "
    #  &#39; gets converted back to '
    # So we need a custom function here, instead of django.utils.html.escape
    return text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')


class FuncSeleniumMixin(CommonMixin):

    @classmethod
    def setUpClass(cls):
        if not cls.display_browser_window():
            cls.__display = Display(visible=False)
            cls.__display.start()
        driver_name = cls.get_driver_name()
        kwargs = {}
        cls._driver = getattr(webdriver, driver_name)(**kwargs)
        cls._driver.set_page_load_timeout(cls.get_page_load_timeout())
        super(FuncSeleniumMixin, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        if not cls.display_browser_window():
            cls._driver.quit()
            cls.__display.stop()
        super(FuncSeleniumMixin, cls).tearDownClass()

    def setUp(self):
        self._have_visited_page = False
        super(FuncSeleniumMixin, self).setUp()
        size = self.get_browser_window_size()
        if size is not None:
            self.set_window_size(*size)
        self._driver.delete_all_cookies()

    # Common API:
    is_full_browser_test = True

    def assertTextPresent(self, text):
        self.assertIn(escape_selenium(text), self._get_page_source())

    def assertTextAbsent(self, text):
        self.assertNotIn(escape_selenium(text), self._get_page_source())

    def back(self):
        self._driver.back()

    @property
    def current_url(self):
        return self._driver.current_url

    def follow_link(self, css_selector):
        return self.click(css_selector, wait_for_reload=True)

    def fill(self, fields):
        for k, v in fields.items():
            e = self._find_with_timeout(css_selector=k)
            self._fill_input(e, v)

    def fill_by_text(self, fields):
        for selector, text in fields.items():
            elem = self._find_with_timeout(css_selector=selector)
            self._fill_input_by_text(elem, text)

    def get_url(self, name, *args, **kwargs):
        """
        Gets the named URL, passing *args and **kwargs to Django's URL 'reverse' function.
        """
        kwargs.pop('expect_errors', None)
        self.get_literal_url(reverse(name, args=args, kwargs=kwargs))
        self.wait_until_loaded('body')
        # TODO - need tests
        # self.wait_for_ajax()

    def get_literal_url(self, url, auto_follow=None, expect_errors=None):
        """
        Gets the passed in URL, as a literal relative URL, without using reverse.
        """
        self._get_url_raw(self.live_server_url + url)

    def is_element_present(self, css_selector):
        try:
            self._driver.find_element_by_css_selector(css_selector)
        except NoSuchElementException:
            return False
        return True

    def set_session_data(self, item_dict):
        # Cookies don't work unless we visit a page first
        if not self._have_visited_page:
            self.get_url('django_functest.emptypage')

        session = self._get_session()
        for name, value in item_dict.items():
            session[name] = text_type(value)
        session.save()

        s2 = self._get_session()
        if all(s2.get(name) == text_type(value) for name, value in item_dict.items()):
            return

        raise RuntimeError("Session not saved correctly")

    def submit(self, css_selector, wait_for_reload=True, auto_follow=None, window_closes=False):
        self.click(css_selector, wait_for_reload=wait_for_reload, window_closes=window_closes)

    def value(self, css_selector):
        elem = self._find(css_selector=css_selector)
        if elem.tag_name == 'input' and elem.get_attribute('type') == 'checkbox':
            return self._is_checked(elem)
        elif elem.tag_name == 'textarea':
            return elem.text
        else:
            return elem.get_attribute('value')

    # Full browser specific:

    # Configuration methods and attributes
    browser_window_size = None

    display = False  # Display browser window or not?

    default_timeout = 10  # seconds

    driver_name = "Firefox"  # Sensible default, works most places

    page_load_timeout = 20  # seconds

    def get_browser_window_size(self):
        return self.browser_window_size

    @classmethod
    def display_browser_window(cls):
        return cls.display

    @classmethod
    def get_default_timeout(cls):
        return cls.default_timeout

    @classmethod
    def get_driver_name(cls):
        return cls.driver_name

    @classmethod
    def get_page_load_timeout(cls):
        return cls.page_load_timeout

    # Runtime methods:

    def click(self, css_selector=None, xpath=None,
              text=None, text_parent_id=None,
              wait_for_reload=False,
              wait_timeout=None,
              double=False, scroll=True, window_closes=False):
        if window_closes:
            wait_for_reload = False
        if wait_for_reload:
            self._driver.execute_script("document.pageReloadedYetFlag='notyet';")

        elem = self._find_with_timeout(css_selector=css_selector,
                                       xpath=xpath,
                                       text=text,
                                       text_parent_id=text_parent_id,
                                       timeout=wait_timeout)
        if scroll:
            self._scroll_into_view(elem)
        elem.click()
        if double:
            try:
                elem.click()
            except StaleElementReferenceException:
                pass

        if wait_for_reload:
            def f(driver):
                obj = driver.execute_script("return document.pageReloadedYetFlag;")

                if obj is None or obj != "notyet":
                    return True
                return False
            try:
                WebDriverWait(self._driver, self.get_default_timeout()).until(f)
            except NoSuchWindowException:
                # legitimate sometimes e.g. when window closes
                pass
        if not window_closes:
            self._wait_until_finished()

    def execute_script(self, script, *args):
        return self._driver.execute_script(script, *args)

    def hover(self, css_selector):
        elem = self._find(css_selector=css_selector)
        self._scroll_into_view(elem)
        ActionChains(self._driver).move_to_element(elem).perform()

    def is_element_displayed(self, css_selector):
        try:
            elem = self._driver.find_element_by_css_selector(css_selector)
        except NoSuchElementException:
            return False
        return elem.is_displayed()

    def save_screenshot(self, dirname="./", filename=None):
        # Especially useful when hiding the browser window gives different behaviour.
        testname = '%s.%s.%s' % (self.__class__.__module__, self.__class__.__name__, self._testMethodName)
        if filename is None:
            filename = datetime.now().strftime('Screenshot %Y-%m-%d %H.%M.%S') + " " + testname + ".png"
        name = os.path.abspath(os.path.join(dirname, filename))
        self._driver.save_screenshot(name)
        return name

    def set_window_size(self, width, height):
        def f(driver):
            driver.set_window_size(width, height)
            win_width, win_height = self._get_window_size()
            # Some drivers fail to get it exactly
            return ((width - 2 <= win_width <= width + 2) and
                    (height - 2 <= win_height <= height + 2))
        self.wait_until(f)

    def switch_window(self, handle=None):
        """
        Switches window.

        If there are only 2 windows, it can work out which window to switch to.
        Otherwise, you should pass in the window handle as `handle` kwarg.

        Returns a tuple (old_window_handle, new_window_handle)
        """
        try:
            current_window_handle = self._driver.current_window_handle
        except NoSuchWindowException:
            # Window closed recently
            current_window_handle = None
        window_handles = self._driver.window_handles
        if handle is None:
            possible_window_handles = [h for h in window_handles
                                       if h != current_window_handle]

            if len(possible_window_handles) > 1:
                raise AssertionError("Don't know which window to switch to!")
            else:
                handle = possible_window_handles[0]

        def f(driver):
            driver.switch_to_window(handle)
            return driver.current_window_handle == handle
        self.wait_until(f)
        return current_window_handle, handle

    def wait_for_page_load(self):
        self.wait_until_loaded('body')
        self._wait_for_document_ready()

    def wait_until(self, callback, timeout=None):
        """
        Helper function that blocks the execution of the tests until the
        specified callback returns a value that is not falsy. This function can
        be called, for example, after clicking a link or submitting a form.
        See the other public methods that call this function for more details.
        """
        if timeout is None:
            timeout = self.get_default_timeout()
        WebDriverWait(self._driver, timeout).until(callback)

    def wait_until_loaded(self, css_selector=None, xpath=None,
                          text=None, text_parent_id=None,
                          timeout=None):
        """
        Helper function that blocks until the element with the given tag name
        is found on the page.
        """
        self.wait_until(self._get_finder(css_selector=css_selector, xpath=xpath,
                                         text=text, text_parent_id=text_parent_id),
                        timeout=timeout)

    # Implementation methods - private

    def _get_finder(self, css_selector=None, xpath=None, text=None, text_parent_id=None):
        if css_selector is not None:
            return lambda driver: driver.find_element_by_css_selector(css_selector)
        if xpath is not None:
            return lambda driver: driver.find_element_by_xpath(xpath)
        if text is not None:
            if text_parent_id is not None:
                prefix = '//*[@id="{0}"]'.format(text_parent_id)
            else:
                prefix = ''
            _xpath = prefix + '//*[contains(text(), "{0}")]'.format(text)
            return lambda driver: driver.find_element_by_xpath(_xpath)
        raise AssertionError("No selector passed in")

    def _get_url_raw(self, url):
        """
        'raw' method for getting URL - not compatible between FullBrowserTest and WebTestBase
        """
        self._have_visited_page = True
        self._driver.get(url)

    def _add_cookie(self, cookie_dict):
        self._driver.add_cookie(cookie_dict)

    def _get_session(self):
        session_cookie = self._driver.get_cookie(settings.SESSION_COOKIE_NAME)
        if session_cookie is None:
            # Create new
            session = get_session_store()
            cookie_data = {'name': settings.SESSION_COOKIE_NAME,
                           'value': session.session_key,
                           'path': '/',
                           'secure': False,
                           }
            if self._driver.name == 'phantomjs':
                # Not sure why this is needed, but it seems to do the trick
                cookie_data['domain'] = '.localhost'

            self._add_cookie(cookie_data)
        else:
            session = get_session_store(session_key=session_cookie['value'])
        return session

    def _get_window_size(self):
        if self._driver.name == "phantomjs":
            return self.execute_script("return [document.width, document.height]")
        else:
            return self.execute_script("return [window.outerWidth, window.outerHeight]")

    def _wait_for_document_ready(self):
        self.wait_until(lambda driver: driver.execute_script("return document.readyState") == "complete")

    def _wait_until_finished(self):
        try:
            self.wait_for_page_load()
        except NoSuchWindowException:
            pass  # window can legitimately close e.g. for popups

    def _get_page_source(self):
        return self._driver.page_source

    def _fill_input(self, elem, val):
        if elem.tag_name == 'select':
            self._set_select_elem(elem, val)
        elif elem.tag_name == 'input' and elem.get_attribute('type') == 'checkbox':
            self._set_check_box(elem, val)
        else:
            self._scroll_into_view(elem)
            elem.clear()
            elem.send_keys(val)

    def _fill_input_by_text(self, elem, val):
        if elem.tag_name == 'select':
            self._set_select_elem_by_text(elem, val)
        else:
            raise SeleniumCantUseElement("Can't do 'fill_by_text' on elements of type {0}".format(elem.tag_name))

    def _find(self, css_selector=None, xpath=None, text=None, text_parent_id=None):
        return self._get_finder(css_selector=css_selector, xpath=xpath,
                                text=text, text_parent_id=text_parent_id)(self._driver)

    def _find_with_timeout(self, css_selector=None, xpath=None, text=None, text_parent_id=None, timeout=None):
        if timeout != 0:
            self.wait_until_loaded(css_selector=css_selector, xpath=xpath,
                                   text=text, text_parent_id=text_parent_id,
                                   timeout=timeout)
        return self._find(css_selector=css_selector, xpath=xpath,
                          text=text, text_parent_id=text_parent_id)

    def _scroll_into_view(self, elem, attempts=0):
        if self._is_visible(elem):
            return

        # Attempt to scroll to the center of the screen. This is the best
        # location to avoid fixed navigation bars which tend to be at the
        # top and bottom.
        viewport_width, viewport_height, doc_width, doc_height = self._scroll_center_data()
        elem_x, elem_y = elem.location['x'], elem.location['y']
        elem_w, elem_h = elem.size['width'], elem.size['height']
        scroll_to_x = elem_x + elem_w / 2 - viewport_width / 2
        scroll_to_y = elem_y + elem_h / 2 - viewport_height / 2

        def clip(val, min_val, max_val):
            return max(min(val, max_val), min_val)

        scroll_to_x = clip(scroll_to_x, 0, doc_width)
        scroll_to_y = clip(scroll_to_y, 0, doc_height)

        self._driver.execute_script("window.scrollTo({0}, {1});".format(
            scroll_to_x, scroll_to_y))
        x, y = self._scroll_position()
        if (x, y) != (scroll_to_x, scroll_to_y):
            # Probably in the middle of another scroll. Wait and try again.
            if attempts < 10:
                time.sleep(0.1)
                self._scroll_into_view(elem, attempts=attempts + 1)
            else:
                logger.warning("Can't scroll to (%s, %s)", scroll_to_x, scroll_to_y)

        if attempts == 0:
            self.wait_until(lambda *_: self._is_visible(elem))

    def _scroll_center_data(self):
        # http://www.howtocreate.co.uk/tutorials/javascript/browserwindow
        return self.execute_script("""return [window.innerWidth,
                                              window.innerHeight,
                                              document.documentElement.offsetWidth,
                                              document.documentElement.offsetHeight];""")

    def _scroll_position(self):
        return self.execute_script("""return [document.documentElement.scrollTop,
                                              document.documentElement.scrollLeft];""")

    def _is_visible(self, elem):
        # Thanks http://stackoverflow.com/a/15203639/182604
        return self.execute_script("""return (function (el) {
    var rect     = el.getBoundingClientRect(),
        vWidth   = window.innerWidth || doc.documentElement.clientWidth,
        vHeight  = window.innerHeight || doc.documentElement.clientHeight,
        efp      = function (x, y) { return document.elementFromPoint(x, y) };

    // Return false if it's not in the viewport
    if (rect.right < 0 || rect.bottom < 0
            || rect.left > vWidth || rect.top > vHeight)
        return false;

    // Return true if any of its four corners or center are visible
    return (
          el.contains(efp(rect.left,  rect.top))
      ||  el.contains(efp(rect.right, rect.top))
      ||  el.contains(efp(rect.right, rect.bottom))
      ||  el.contains(efp(rect.left,  rect.bottom))
      ||  el.contains(efp(rect.right - rect.width / 2, rect.bottom - rect.height / 2))
    );
})(arguments[0])""", elem)

    def _is_checked(self, elem):
        return elem.get_attribute('checked') == 'true'

    def _set_check_box(self, elem, state):
        if self._is_checked(elem) != state:
            self._scroll_into_view(elem)
            elem.click()

    def _set_select_elem(self, elem, value):
        self._scroll_into_view(elem)
        s = Select(elem)
        s.select_by_value(value)

    def _set_select_elem_by_text(self, elem, text):
        self._scroll_into_view(elem)
        s = Select(elem)
        s.select_by_visible_text(text)
